@@ -31,13 +31,9 @@
 
 #include <glib/gstdio.h>
 
-#include <libtracker-common/tracker-file-utils.h>
-#include <libtracker-common/tracker-parser.h>
-#include <libtracker-common/tracker-type-utils.h>
-#include <libtracker-common/tracker-utils.h>
-#include <libtracker-common/tracker-ontology.h>
+#include <libtracker-common/tracker-common.h>
 
-#include <tracker-fts/tracker-fts.h>
+#include <libtracker-fts/tracker-fts.h>
 
 #include <libtracker-db/tracker-db-interface-sqlite.h>
 #include <libtracker-db/tracker-db-manager.h>
@@ -58,6 +54,7 @@
 #define RDFS_SUB_PROPERTY_OF RDFS_PREFIX "subPropertyOf"
 
 #define NRL_PREFIX TRACKER_NRL_PREFIX
+#define NRL_INVERSE_FUNCTIONAL_PROPERTY TRACKER_NRL_PREFIX "InverseFunctionalProperty"
 #define NRL_MAX_CARDINALITY NRL_PREFIX "maxCardinality"
 
 #define TRACKER_PREFIX TRACKER_TRACKER_PREFIX
@@ -73,7 +70,13 @@ load_ontology_file_from_path (const gchar	 *ontology_file)
 	TrackerTurtleReader *reader;
 	GError              *error = NULL;
 
-	reader = tracker_turtle_reader_new (ontology_file);
+	reader = tracker_turtle_reader_new (ontology_file, &error);
+	if (error) {
+		g_critical ("Turtle parse error: %s", error->message);
+		g_error_free (error);
+		return;
+	}
+
 	while (error == NULL && tracker_turtle_reader_next (reader, &error)) {
 		const gchar *subject, *predicate, *object;
 
@@ -106,6 +109,16 @@ load_ontology_file_from_path (const gchar	 *ontology_file)
 				tracker_property_set_uri (property, subject);
 				tracker_ontology_add_property (property);
 				g_object_unref (property);
+			} else if (g_strcmp0 (object, NRL_INVERSE_FUNCTIONAL_PROPERTY) == 0) {
+				TrackerProperty *property;
+
+				property = tracker_ontology_get_property_by_uri (subject);
+				if (property == NULL) {
+					g_critical ("%s: Unknown property %s", ontology_file, subject);
+					continue;
+				}
+
+				tracker_property_set_is_inverse_functional_property (property, TRUE);
 			} else if (g_strcmp0 (object, TRACKER_PREFIX "Namespace") == 0) {
 				TrackerNamespace *namespace;
 
@@ -221,6 +234,18 @@ load_ontology_file_from_path (const gchar	 *ontology_file)
 			if (g_strcmp0 (object, "true") == 0) {
 				tracker_property_set_transient (property, TRUE);
 			}
+		} else if (g_strcmp0 (predicate, TRACKER_PREFIX "isAnnotation") == 0) {
+			TrackerProperty *property;
+
+			property = tracker_ontology_get_property_by_uri (subject);
+			if (property == NULL) {
+				g_critical ("%s: Unknown property %s", ontology_file, subject);
+				continue;
+			}
+
+			if (g_strcmp0 (object, "true") == 0) {
+				tracker_property_set_embedded (property, FALSE);
+			}
 		} else if (g_strcmp0 (predicate, TRACKER_PREFIX "fulltextIndexed") == 0) {
 			TrackerProperty *property;
 
@@ -372,7 +397,7 @@ db_get_static_data (TrackerDBInterface *iface)
 	}
 
 	stmt = tracker_db_interface_create_statement (iface,
-						      "SELECT (SELECT Uri FROM \"rdfs:Resource\" WHERE ID = \"rdfs:Class\".ID) "
+						      "SELECT \"rdfs:Class\".ID, (SELECT Uri FROM \"rdfs:Resource\" WHERE ID = \"rdfs:Class\".ID) "
 						      "FROM \"rdfs:Class\" ORDER BY ID");
 	cursor = tracker_db_statement_start_cursor (stmt, NULL);
 	g_object_unref (stmt);
@@ -381,15 +406,19 @@ db_get_static_data (TrackerDBInterface *iface)
 		while (tracker_db_cursor_iter_next (cursor)) {
 			TrackerClass *class;
 			const gchar  *uri;
+			gint          id;
 			gint          count;
 
 			class = tracker_class_new ();
 
-			uri = tracker_db_cursor_get_string (cursor, 0);
+			id = tracker_db_cursor_get_int (cursor, 0);
+			uri = tracker_db_cursor_get_string (cursor, 1);
 
 			tracker_class_set_uri (class, uri);
 			class_add_super_classes_from_db (iface, class);
+
 			tracker_ontology_add_class (class);
+			tracker_ontology_add_id_uri_pair (id, uri);
 
 			/* xsd classes do not derive from rdfs:Resource and do not use separate tables */
 			if (!g_str_has_prefix (tracker_class_get_name (class), "xsd:")) {
@@ -409,13 +438,16 @@ db_get_static_data (TrackerDBInterface *iface)
 	}
 
 	stmt = tracker_db_interface_create_statement (iface,
-						      "SELECT (SELECT Uri FROM \"rdfs:Resource\" WHERE ID = \"rdf:Property\".ID), "
+						      "SELECT \"rdf:Property\".ID, (SELECT Uri FROM \"rdfs:Resource\" WHERE ID = \"rdf:Property\".ID), "
 						      "(SELECT Uri FROM \"rdfs:Resource\" WHERE ID = \"rdfs:domain\"), "
 						      "(SELECT Uri FROM \"rdfs:Resource\" WHERE ID = \"rdfs:range\"), "
 						      "\"nrl:maxCardinality\", "
 						      "\"tracker:indexed\", "
 						      "\"tracker:fulltextIndexed\", "
-						      "\"tracker:transient\" "
+						      "\"tracker:transient\", "
+						      "\"tracker:isAnnotation\", "
+						      "(SELECT 1 FROM \"rdfs:Resource_rdf:type\" WHERE ID = \"rdf:Property\".ID AND "
+						              "\"rdf:type\" = (Select ID FROM \"rdfs:Resource\" WHERE Uri = '" NRL_INVERSE_FUNCTIONAL_PROPERTY "')) "
 						      "FROM \"rdf:Property\" ORDER BY ID");
 	cursor = tracker_db_statement_start_cursor (stmt, NULL);
 	g_object_unref (stmt);
@@ -427,15 +459,17 @@ db_get_static_data (TrackerDBInterface *iface)
 			TrackerProperty *property;
 			const gchar     *uri, *domain_uri, *range_uri;
 			gboolean         multi_valued, indexed, fulltext_indexed;
-			gboolean         transient = FALSE;
+			gboolean         transient, annotation, is_inverse_functional_property;
+			gint             id;
 
 			property = tracker_property_new ();
 
-			uri = tracker_db_cursor_get_string (cursor, 0);
-			domain_uri = tracker_db_cursor_get_string (cursor, 1);
-			range_uri = tracker_db_cursor_get_string (cursor, 2);
+			id = tracker_db_cursor_get_int (cursor, 0);
+			uri = tracker_db_cursor_get_string (cursor, 1);
+			domain_uri = tracker_db_cursor_get_string (cursor, 2);
+			range_uri = tracker_db_cursor_get_string (cursor, 3);
 
-			tracker_db_cursor_get_value (cursor, 3, &value);
+			tracker_db_cursor_get_value (cursor, 4, &value);
 
 			if (G_VALUE_TYPE (&value) != 0) {
 				multi_valued = (g_value_get_int (&value) > 1);
@@ -446,7 +480,7 @@ db_get_static_data (TrackerDBInterface *iface)
 				multi_valued = TRUE;
 			}
 
-			tracker_db_cursor_get_value (cursor, 4, &value);
+			tracker_db_cursor_get_value (cursor, 5, &value);
 
 			if (G_VALUE_TYPE (&value) != 0) {
 				indexed = (g_value_get_int (&value) == 1);
@@ -456,7 +490,7 @@ db_get_static_data (TrackerDBInterface *iface)
 				indexed = FALSE;
 			}
 
-			tracker_db_cursor_get_value (cursor, 5, &value);
+			tracker_db_cursor_get_value (cursor, 6, &value);
 
 			if (G_VALUE_TYPE (&value) != 0) {
 				fulltext_indexed = (g_value_get_int (&value) == 1);
@@ -466,7 +500,7 @@ db_get_static_data (TrackerDBInterface *iface)
 				fulltext_indexed = FALSE;
 			}
 
-			tracker_db_cursor_get_value (cursor, 6, &value);
+			tracker_db_cursor_get_value (cursor, 7, &value);
 
 			if (G_VALUE_TYPE (&value) != 0) {
 				transient = (g_value_get_int (&value) == 1);
@@ -476,6 +510,26 @@ db_get_static_data (TrackerDBInterface *iface)
 				transient = FALSE;
 			}
 
+			tracker_db_cursor_get_value (cursor, 8, &value);
+
+			if (G_VALUE_TYPE (&value) != 0) {
+				annotation = (g_value_get_int (&value) == 1);
+				g_value_unset (&value);
+			} else {
+				/* NULL */
+				annotation = FALSE;
+			}
+
+			tracker_db_cursor_get_value (cursor, 8, &value);
+
+			if (G_VALUE_TYPE (&value) != 0) {
+				is_inverse_functional_property = TRUE;
+				g_value_unset (&value);
+			} else {
+				/* NULL */
+				is_inverse_functional_property = FALSE;
+			}
+
 			tracker_property_set_transient (property, transient);
 			tracker_property_set_uri (property, uri);
 			tracker_property_set_domain (property, tracker_ontology_get_class_by_uri (domain_uri));
@@ -483,8 +537,12 @@ db_get_static_data (TrackerDBInterface *iface)
 			tracker_property_set_multiple_values (property, multi_valued);
 			tracker_property_set_indexed (property, indexed);
 			tracker_property_set_fulltext_indexed (property, fulltext_indexed);
+			tracker_property_set_embedded (property, !annotation);
+			tracker_property_set_is_inverse_functional_property (property, is_inverse_functional_property);
 			property_add_super_properties_from_db (iface, property);
+
 			tracker_ontology_add_property (property);
+			tracker_ontology_add_id_uri_pair (id, uri);
 
 			g_object_unref (property);
 
@@ -494,11 +552,41 @@ db_get_static_data (TrackerDBInterface *iface)
 	}
 }
 
+
+static void
+insert_uri_in_resource_table (TrackerDBInterface *iface, const gchar *uri, gint *max_id)
+{
+	TrackerDBStatement *stmt;
+	gint id = ++(*max_id);
+	GError *error = NULL;
+
+	stmt = tracker_db_interface_create_statement (iface,
+	                                              "INSERT  "
+	                                              "INTO \"rdfs:Resource\" "
+	                                              "(ID, Uri, \"tracker:added\", "
+	                                                "\"tracker:modified\", Available) "
+	                                              "VALUES (?, ?, ?, 0, 1)");
+	tracker_db_statement_bind_int (stmt, 0, id);
+	tracker_db_statement_bind_text (stmt, 1, uri);
+	tracker_db_statement_bind_int64 (stmt, 2, (gint64) time (NULL));
+	tracker_db_statement_execute (stmt, &error);
+
+	if (error) {
+		g_critical ("%s\n", error->message);
+		g_clear_error (&error);
+	}
+
+	tracker_ontology_add_id_uri_pair (id, uri);
+	g_object_unref (stmt);
+
+}
+
 static void
 create_decomposed_metadata_property_table (TrackerDBInterface *iface, 
 					   TrackerProperty   **property, 
 					   const gchar        *service_name,
-					   const gchar       **sql_type_for_single_value)
+					   const gchar       **sql_type_for_single_value,
+					   gint               *max_id)
 {
 	const char *field_name;
 	const char *sql_type;
@@ -526,9 +614,6 @@ create_decomposed_metadata_property_table (TrackerDBInterface *iface,
 	case TRACKER_PROPERTY_TYPE_DOUBLE:
 		sql_type = "REAL";
 		break;
-	case TRACKER_PROPERTY_TYPE_BLOB:
-	case TRACKER_PROPERTY_TYPE_STRUCT:
-	case TRACKER_PROPERTY_TYPE_FULLTEXT:
 	default:
 		sql_type = "";
 		break;
@@ -576,6 +661,11 @@ create_decomposed_metadata_property_table (TrackerDBInterface *iface,
 		*sql_type_for_single_value = sql_type;
 	}
 
+	/* insert property uri in rdfs:Resource table */
+	if (max_id && tracker_property_get_uri (*property) != NULL) {
+		insert_uri_in_resource_table (iface, tracker_property_get_uri (*property),
+		                              max_id);
+	}
 }
 
 static void
@@ -611,7 +701,8 @@ create_decomposed_metadata_tables (TrackerDBInterface *iface,
 
 			create_decomposed_metadata_property_table (iface, property, 
 								   service_name, 
-								   &sql_type_for_single_value);
+								   &sql_type_for_single_value,
+								   main_class ? NULL : max_id);
 
 			if (sql_type_for_single_value) {
 				/* single value */
@@ -621,6 +712,9 @@ create_decomposed_metadata_tables (TrackerDBInterface *iface,
 				g_string_append_printf (sql, ", \"%s\" %s", 
 							tracker_property_get_name (*property), 
 							sql_type_for_single_value);
+				if (tracker_property_get_is_inverse_functional_property (*property)) {
+					g_string_append (sql, " UNIQUE");
+				}
 			}
 		}
 	}
@@ -652,20 +746,14 @@ create_decomposed_metadata_tables (TrackerDBInterface *iface,
 		}
 	}
 
-	g_slist_free (class_properties);
-
 	/* insert class uri in rdfs:Resource table */
 	if (tracker_class_get_uri (service) != NULL) {
-		TrackerDBStatement *stmt;
-
-		stmt = tracker_db_interface_create_statement (iface,
-							      "INSERT OR IGNORE INTO \"rdfs:Resource\" (ID, Uri, \"tracker:added\", \"tracker:modified\") VALUES (?, ?, ?, 0)");
-		tracker_db_statement_bind_int (stmt, 0, ++(*max_id));
-		tracker_db_statement_bind_text (stmt, 1, tracker_class_get_uri (service));
-		tracker_db_statement_bind_int64 (stmt, 2, (gint64) time (NULL));
-		tracker_db_statement_execute (stmt, NULL);
-		g_object_unref (stmt);
+		insert_uri_in_resource_table (iface, tracker_class_get_uri (service),
+		                              max_id);
 	}
+
+	g_slist_free (class_properties);
+
 }
 
 static void
@@ -691,8 +779,7 @@ create_decomposed_transient_metadata_tables (TrackerDBInterface *iface)
 			/* create the TEMPORARY table */
 			create_decomposed_metadata_property_table (iface, property,
 								   service_name,
-								   NULL);
-			
+								   NULL, NULL);
 		}
 	}
 }
@@ -708,14 +795,23 @@ create_fts_table (TrackerDBInterface *iface)
 }
 
 gboolean
-tracker_data_manager_init (TrackerDBManagerFlags       flags,
-			   const gchar                *test_schema,
-			   gboolean                   *first_time,
-			   gboolean                   *need_journal)
+tracker_data_manager_init (TrackerDBManagerFlags  flags,
+			   const gchar           *test_schema,
+			   gboolean              *first_time,
+			   gboolean              *need_journal)
 {
 	TrackerDBInterface *iface;
 	gboolean is_first_time_index;
 
+	/* First set defaults for return values */
+	if (first_time) {
+		*first_time = FALSE;
+	}
+
+	if (need_journal) {
+		*need_journal = FALSE;
+	}
+
 	if (initialized) {
 		return TRUE;
 	}
