@@ -79,7 +79,10 @@ struct _TrackerDataUpdateBufferProperty {
 
 struct _TrackerDataUpdateBufferTable {
 	gboolean insert;
+	gboolean delete_row;
+	gboolean delete_value;
 	gboolean multiple_values;
+	TrackerClass *class;
 	GArray *properties;
 };
 
@@ -104,6 +107,8 @@ static TrackerStatementCallback delete_callback = NULL;
 static gpointer delete_data;
 static TrackerCommitCallback commit_callback = NULL;
 static gpointer commit_data;
+static TrackerCommitCallback rollback_callback = NULL;
+static gpointer rollback_data;
 
 void 
 tracker_data_set_commit_statement_callback (TrackerCommitCallback    callback,
@@ -113,6 +118,14 @@ tracker_data_set_commit_statement_callback (TrackerCommitCallback    callback,
 	commit_data = user_data;
 }
 
+void
+tracker_data_set_rollback_statement_callback (TrackerCommitCallback    callback,
+					      gpointer                 user_data)
+{
+	rollback_callback = callback;
+	rollback_data = user_data;
+}
+
 void
 tracker_data_set_insert_statement_callback (TrackerStatementCallback callback,
 					    gpointer                 user_data)
@@ -236,11 +249,12 @@ cache_ensure_table (const gchar            *table_name,
 }
 
 static void
-cache_insert_row (const gchar            *table_name)
+cache_insert_row (TrackerClass *class)
 {
 	TrackerDataUpdateBufferTable    *table;
 
-	table = cache_ensure_table (table_name, FALSE);
+	table = cache_ensure_table (tracker_class_get_name (class), FALSE);
+	table->class = class;
 	table->insert = TRUE;
 }
 
@@ -262,6 +276,35 @@ cache_insert_value (const gchar            *table_name,
 	g_array_append_val (table->properties, property);
 }
 
+static void
+cache_delete_row (TrackerClass *class)
+{
+	TrackerDataUpdateBufferTable    *table;
+
+	table = cache_ensure_table (tracker_class_get_name (class), FALSE);
+	table->class = class;
+	table->delete_row = TRUE;
+}
+
+static void
+cache_delete_value (const gchar            *table_name,
+		    const gchar            *field_name,
+		    GValue                 *value,
+		    gboolean                multiple_values,
+		    gboolean                fts)
+{
+	TrackerDataUpdateBufferTable    *table;
+	TrackerDataUpdateBufferProperty  property;
+
+	property.name = g_strdup (field_name);
+	property.value = *value;
+	property.fts = fts;
+
+	table = cache_ensure_table (table_name, multiple_values);
+	table->delete_value = TRUE;
+	g_array_append_val (table->properties, property);
+}
+
 static guint32
 query_resource_id (const gchar *uri)
 {
@@ -335,7 +378,7 @@ statement_bind_gvalue (TrackerDBStatement *stmt,
 }
 
 static void
-tracker_data_resource_buffer_flush (void)
+tracker_data_resource_buffer_flush (GError **error)
 {
 	TrackerDBInterface             *iface;
 	TrackerDBStatement             *stmt;
@@ -345,6 +388,7 @@ tracker_data_resource_buffer_flush (void)
 	const gchar                    *table_name;
 	GString                        *sql, *fts;
 	int                             i;
+	GError                         *actual_error = NULL;
 
 	iface = tracker_db_manager_get_db_interface ();
 
@@ -354,11 +398,16 @@ tracker_data_resource_buffer_flush (void)
 			"UPDATE \"rdfs:Resource\" SET Uri = ? WHERE ID = ?");
 		tracker_db_statement_bind_text (stmt, 0, resource_buffer->new_subject);
 		tracker_db_statement_bind_int (stmt, 1, resource_buffer->id);
-		tracker_db_statement_execute (stmt, NULL);
+		tracker_db_statement_execute (stmt, &actual_error);
 		g_object_unref (stmt);
 
 		g_free (resource_buffer->new_subject);
 		resource_buffer->new_subject = NULL;
+
+		if (actual_error) {
+			g_propagate_error (error, actual_error);
+			return;
+		}
 	}
 
 	g_hash_table_iter_init (&iter, resource_buffer->tables);
@@ -367,24 +416,70 @@ tracker_data_resource_buffer_flush (void)
 			for (i = 0; i < table->properties->len; i++) {
 				property = &g_array_index (table->properties, TrackerDataUpdateBufferProperty, i);
 
-				stmt = tracker_db_interface_create_statement (iface,
-					"INSERT OR IGNORE INTO \"%s\" (ID, \"%s\") VALUES (?, ?)",
-					table_name,
-					property->name);
+				if (table->delete_value) {
+					/* delete rows for multiple value properties */
+					stmt = tracker_db_interface_create_statement (iface,
+						"DELETE FROM \"%s\" WHERE ID = ? AND \"%s\" = ?",
+						table_name,
+						property->name);
+				} else {
+					stmt = tracker_db_interface_create_statement (iface,
+						"INSERT OR IGNORE INTO \"%s\" (ID, \"%s\") VALUES (?, ?)",
+						table_name,
+						property->name);
+				}
+
 				tracker_db_statement_bind_int (stmt, 0, resource_buffer->id);
 				statement_bind_gvalue (stmt, 1, &property->value);
-				tracker_db_statement_execute (stmt, NULL);
+				tracker_db_statement_execute (stmt, &actual_error);
 				g_object_unref (stmt);
+
+				if (actual_error) {
+					g_propagate_error (error, actual_error);
+					return;
+				}
 			}
 		} else {
+			if (table->delete_row) {
+				/* remove entry from rdf:type table */
+				stmt = tracker_db_interface_create_statement (iface, "DELETE FROM \"rdfs:Resource_rdf:type\" WHERE ID = ? AND \"rdf:type\" = ?");
+				tracker_db_statement_bind_int (stmt, 0, resource_buffer->id);
+				tracker_db_statement_bind_int (stmt, 1, ensure_resource_id (tracker_class_get_uri (table->class)));
+				tracker_db_statement_execute (stmt, &actual_error);
+				g_object_unref (stmt);
+
+				if (actual_error) {
+					g_propagate_error (error, actual_error);
+					return;
+				}
+
+				/* remove row from class table */
+				stmt = tracker_db_interface_create_statement (iface, "DELETE FROM \"%s\" WHERE ID = ?", table_name);
+				tracker_db_statement_bind_int (stmt, 0, resource_buffer->id);
+				tracker_db_statement_execute (stmt, &actual_error);
+				g_object_unref (stmt);
+
+				if (actual_error) {
+					g_propagate_error (error, actual_error);
+					return;
+				}
+
+				continue;
+			}
+
 			if (table->insert) {
 				/* ensure we have a row for the subject id */
 				stmt = tracker_db_interface_create_statement (iface,
 					"INSERT OR IGNORE INTO \"%s\" (ID) VALUES (?)",
 					table_name);
 				tracker_db_statement_bind_int (stmt, 0, resource_buffer->id);
-				tracker_db_statement_execute (stmt, NULL);
+				tracker_db_statement_execute (stmt, &actual_error);
 				g_object_unref (stmt);
+
+				if (actual_error) {
+					g_propagate_error (error, actual_error);
+					return;
+				}
 			}
 
 			if (table->properties->len == 0) {
@@ -409,13 +504,23 @@ tracker_data_resource_buffer_flush (void)
 
 			for (i = 0; i < table->properties->len; i++) {
 				property = &g_array_index (table->properties, TrackerDataUpdateBufferProperty, i);
-				statement_bind_gvalue (stmt, i, &property->value);
+				if (table->delete_value) {
+					/* just set value to NULL for single value properties */
+					tracker_db_statement_bind_null (stmt, i);
+				} else {
+					statement_bind_gvalue (stmt, i, &property->value);
+				}
 			}
 
-			tracker_db_statement_execute (stmt, NULL);
+			tracker_db_statement_execute (stmt, &actual_error);
 			g_object_unref (stmt);
 
 			g_string_free (sql, TRUE);
+
+			if (actual_error) {
+				g_propagate_error (error, actual_error);
+				return;
+			}
 		}
 	}
 
@@ -458,13 +563,18 @@ static void resource_buffer_free (TrackerDataUpdateBufferResource *resource)
 }
 
 void
-tracker_data_update_buffer_flush (void)
+tracker_data_update_buffer_flush (GError **error)
 {
 	GHashTableIter iter;
+	GError *actual_error = NULL;
 
 	g_hash_table_iter_init (&iter, update_buffer.resources);
 	while (g_hash_table_iter_next (&iter, NULL, (gpointer*) &resource_buffer)) {
-		tracker_data_resource_buffer_flush ();
+		tracker_data_resource_buffer_flush (&actual_error);
+		if (actual_error) {
+			g_propagate_error (error, actual_error);
+			break;
+		}
 	}
 
 	g_hash_table_remove_all (update_buffer.resources);
@@ -561,17 +671,16 @@ cache_create_service_decomposed (TrackerClass           *cl)
 		}
 	}
 
-	tracker_class_set_count (cl, tracker_class_get_count (cl) + 1);
-
 	g_ptr_array_add (resource_buffer->types, g_strdup (class_uri));
 
 	g_value_init (&gvalue, G_TYPE_INT);
 
-	cache_insert_row (tracker_class_get_name (cl));
+	cache_insert_row (cl);
 
 	g_value_set_int (&gvalue, ensure_resource_id (tracker_class_get_uri (cl)));
 	cache_insert_value ("rdfs:Resource_rdf:type", "rdf:type", &gvalue, TRUE, FALSE);
 
+	tracker_class_set_count (cl, tracker_class_get_count (cl) + 1);
 	if (insert_callback) {
 		insert_callback (resource_buffer->subject, RDF_PREFIX "type", class_uri, 
 		                 resource_buffer->types, insert_data);
@@ -621,6 +730,28 @@ value_set_add_value (GValueArray *value_set,
 	return TRUE;
 }
 
+static gboolean
+value_set_remove_value (GValueArray *value_set,
+                        GValue      *value)
+{
+	gint i;
+
+	g_return_val_if_fail (G_VALUE_TYPE (value), FALSE);
+
+	for (i = 0; i < value_set->n_values; i++) {
+		if (value_equal (g_value_array_get_nth (value_set, i), value)) {
+			/* value found, remove from set */
+
+			g_value_array_remove (value_set, i);
+
+			return TRUE;
+		}
+	}
+
+	/* no change, value not found */
+	return FALSE;
+}
+
 static gboolean
 check_property_domain (TrackerProperty *property)
 {
@@ -640,11 +771,10 @@ get_property_values (TrackerProperty *property)
 {
 	TrackerDBInterface *iface;
 	TrackerDBStatement *stmt;
-	TrackerDBCursor    *cursor;
+	TrackerDBResultSet *result_set;
 	gchar              *table_name;
 	const gchar        *field_name;
 	gboolean            multiple_values;
-	GValue gvalue = { 0 };
 	GValueArray        *old_values;
 
 	multiple_values = tracker_property_get_multiple_values (property);
@@ -666,19 +796,29 @@ get_property_values (TrackerProperty *property)
 
 		stmt = tracker_db_interface_create_statement (iface, "SELECT \"%s\" FROM \"%s\" WHERE ID = ?", field_name, table_name);
 		tracker_db_statement_bind_int (stmt, 0, resource_buffer->id);
-		cursor = tracker_db_statement_start_cursor (stmt, NULL);
-		g_object_unref (stmt);
+		result_set = tracker_db_statement_execute (stmt, NULL);
 
-		if (cursor) {
-			while (tracker_db_cursor_iter_next (cursor)) {
-				tracker_db_cursor_get_value (cursor, 0, &gvalue);
+		/* We use a result_set instead of a cursor here because it's
+		 * possible that otherwise the cursor would remain open during
+		 * the call from delete_resource_description. In future we want 
+		 * to allow having the same query open on multiple cursors, 
+		 * right now we don't support this. Which is why this workaround */
+
+		if (result_set) {
+			gboolean valid = TRUE;
+
+			while (valid) {
+				GValue gvalue = { 0 };
+				_tracker_db_result_set_get_value (result_set, 0, &gvalue);
 				if (G_VALUE_TYPE (&gvalue)) {
 					g_value_array_append (old_values, &gvalue);
 					g_value_unset (&gvalue);
 				}
+				valid = tracker_db_result_set_iter_next (result_set);
 			}
-			g_object_unref (cursor);
+			g_object_unref (result_set);
 		}
+		g_object_unref (stmt);
 	}
 
 	g_free (table_name);
@@ -686,39 +826,14 @@ get_property_values (TrackerProperty *property)
 	return old_values;
 }
 
-static void
-cache_set_metadata_decomposed (TrackerProperty	*property,
-			       const gchar	*value,
-			       GError          **error)
+static GValueArray *
+get_old_property_values (TrackerProperty  *property,
+                         GError          **error)
 {
-	guint32		    object_id;
-	gboolean            multiple_values, fts;
-	gchar              *table_name;
-	const gchar        *field_name;
-	TrackerProperty   **properties, **super_properties, **prop;
-	GValue gvalue = { 0 };
+	gboolean            fts;
+	TrackerProperty   **properties, **prop;
 	GValueArray        *old_values;
 
-	/* also insert super property values */
-	super_properties = tracker_property_get_super_properties (property);
-	while (*super_properties) {
-		cache_set_metadata_decomposed (*super_properties, value, error);
-		if (*error) {
-			return;
-		}
-		super_properties++;
-	}
-
-	multiple_values = tracker_property_get_multiple_values (property);
-	if (multiple_values) {
-		table_name = g_strdup_printf ("%s_%s",
-			tracker_class_get_name (tracker_property_get_domain (property)),
-			tracker_property_get_name (property));
-	} else {
-		table_name = g_strdup (tracker_class_get_name (tracker_property_get_domain (property)));
-	}
-	field_name = tracker_property_get_name (property);
-
 	fts = tracker_property_get_fulltext_indexed (property);
 
 	/* read existing property values */
@@ -729,9 +844,8 @@ cache_set_metadata_decomposed (TrackerProperty	*property,
 				     "Subject `%s' is not in domain `%s' of property `%s'",
 				     resource_buffer->subject,
 				     tracker_class_get_name (tracker_property_get_domain (property)),
-				     field_name);
-			g_free (table_name);
-			return;
+				     tracker_property_get_name (property));
+			return NULL;
 		}
 
 		if (fts && !resource_buffer->fts_updated && !resource_buffer->create) {
@@ -769,39 +883,94 @@ cache_set_metadata_decomposed (TrackerProperty	*property,
 		}
 	}
 
-	switch (tracker_property_get_data_type (property)) {
+	return old_values;
+}
+
+static void
+string_to_gvalue (const gchar         *value,
+                  TrackerPropertyType  type,
+                  GValue              *gvalue)
+{
+	guint32		    object_id;
+
+	switch (type) {
 	case TRACKER_PROPERTY_TYPE_STRING:
-		g_value_init (&gvalue, G_TYPE_STRING);
-		g_value_set_string (&gvalue, value);
+		g_value_init (gvalue, G_TYPE_STRING);
+		g_value_set_string (gvalue, value);
 		break;
 	case TRACKER_PROPERTY_TYPE_INTEGER:
-		g_value_init (&gvalue, G_TYPE_INT);
-		g_value_set_int (&gvalue, atoi (value));
+		g_value_init (gvalue, G_TYPE_INT);
+		g_value_set_int (gvalue, atoi (value));
 		break;
 	case TRACKER_PROPERTY_TYPE_BOOLEAN:
 		/* use G_TYPE_INT to be compatible with value stored in DB
 		   (important for value_equal function) */
-		g_value_init (&gvalue, G_TYPE_INT);
-		g_value_set_int (&gvalue, strcmp (value, "true") == 0);
+		g_value_init (gvalue, G_TYPE_INT);
+		g_value_set_int (gvalue, strcmp (value, "true") == 0);
 		break;
 	case TRACKER_PROPERTY_TYPE_DOUBLE:
-		g_value_init (&gvalue, G_TYPE_DOUBLE);
-		g_value_set_double (&gvalue, atof (value));
+		g_value_init (gvalue, G_TYPE_DOUBLE);
+		g_value_set_double (gvalue, atof (value));
 		break;
 	case TRACKER_PROPERTY_TYPE_DATE:
 	case TRACKER_PROPERTY_TYPE_DATETIME:
-		g_value_init (&gvalue, G_TYPE_INT);
-		g_value_set_int (&gvalue, tracker_string_to_date (value));
+		g_value_init (gvalue, G_TYPE_INT);
+		g_value_set_int (gvalue, tracker_string_to_date (value));
 		break;
 	case TRACKER_PROPERTY_TYPE_RESOURCE:
 		object_id = ensure_resource_id (value);
-		g_value_init (&gvalue, G_TYPE_INT);
-		g_value_set_int (&gvalue, object_id);
+		g_value_init (gvalue, G_TYPE_INT);
+		g_value_set_int (gvalue, object_id);
 		break;
 	default:
+		g_warn_if_reached ();
+		break;
+	}
+}
+
+static void
+cache_set_metadata_decomposed (TrackerProperty	*property,
+			       const gchar	*value,
+			       GError          **error)
+{
+	gboolean            multiple_values, fts;
+	gchar              *table_name;
+	const gchar        *field_name;
+	TrackerProperty   **super_properties;
+	GValue gvalue = { 0 };
+	GValueArray        *old_values;
+
+	/* also insert super property values */
+	super_properties = tracker_property_get_super_properties (property);
+	while (*super_properties) {
+		cache_set_metadata_decomposed (*super_properties, value, error);
+		if (*error) {
+			return;
+		}
+		super_properties++;
+	}
+
+	multiple_values = tracker_property_get_multiple_values (property);
+	if (multiple_values) {
+		table_name = g_strdup_printf ("%s_%s",
+			tracker_class_get_name (tracker_property_get_domain (property)),
+			tracker_property_get_name (property));
+	} else {
+		table_name = g_strdup (tracker_class_get_name (tracker_property_get_domain (property)));
+	}
+	field_name = tracker_property_get_name (property);
+
+	fts = tracker_property_get_fulltext_indexed (property);
+
+	/* read existing property values */
+	old_values = get_old_property_values (property, error);
+	if (*error) {
+		g_free (table_name);
 		return;
 	}
 
+	string_to_gvalue (value, tracker_property_get_data_type (property), &gvalue);
+
 	if (!value_set_add_value (old_values, &gvalue)) {
 		/* value already inserted */
 		g_value_unset (&gvalue);
@@ -822,95 +991,153 @@ cache_set_metadata_decomposed (TrackerProperty	*property,
 }
 
 static void
-delete_resource_type (gint resource_id,
-		      TrackerClass           *cl)
+delete_metadata_decomposed (TrackerProperty  *property,
+			    const gchar	     *value,
+			    GError          **error)
 {
-	TrackerDBInterface  *iface;
-	TrackerDBStatement  *stmt;
-
-	iface = tracker_db_manager_get_db_interface ();
-
-	/* remove entry from rdf:type table */
-	stmt = tracker_db_interface_create_statement (iface, "DELETE FROM \"rdfs:Resource_rdf:type\" WHERE ID = ? AND \"rdf:type\" = ?");
-	tracker_db_statement_bind_int (stmt, 0, resource_id);
-	tracker_db_statement_bind_int (stmt, 1, ensure_resource_id (tracker_class_get_uri (cl)));
-	tracker_db_statement_execute (stmt, NULL);
-	g_object_unref (stmt);
-
-	/* remove row from class table */
-	stmt = tracker_db_interface_create_statement (iface, "DELETE FROM \"%s\" WHERE ID = ?", tracker_class_get_name (cl));
-	tracker_db_statement_bind_int (stmt, 0, resource_id);
-	tracker_db_statement_execute (stmt, NULL);
-	g_object_unref (stmt);
-}
-
-static void
-delete_metadata_decomposed (gint resource_id,
-			    TrackerProperty	*property,
-			    const gchar	*value)
-{
-	TrackerDBInterface  *iface;
-	TrackerDBStatement  *stmt;
-	guint32		    object_id;
-	gboolean            multiple_values;
-	const gchar *class_name, *property_name;
+	gboolean            multiple_values, fts;
+	gchar              *table_name;
+	const gchar        *field_name;
 	TrackerProperty   **super_properties;
-
-	iface = tracker_db_manager_get_db_interface ();
+	GValue gvalue = { 0 };
+	GValueArray        *old_values;
 
 	multiple_values = tracker_property_get_multiple_values (property);
-	class_name = tracker_class_get_name (tracker_property_get_domain (property));
-	property_name = tracker_property_get_name (property);
 	if (multiple_values) {
-		/* delete rows for multiple value properties */
-		stmt = tracker_db_interface_create_statement (iface,
-			"DELETE FROM \"%s_%s\" WHERE ID = ? AND \"%s\" = ?",
-			class_name, property_name, property_name);
+		table_name = g_strdup_printf ("%s_%s",
+			tracker_class_get_name (tracker_property_get_domain (property)),
+			tracker_property_get_name (property));
 	} else {
-		/* just set value to NULL for single value properties */
-		stmt = tracker_db_interface_create_statement (iface,
-			"UPDATE \"%s\" SET \"%s\" = NULL WHERE ID = ? AND \"%s\" = ?",
-			class_name, property_name, property_name);
+		table_name = g_strdup (tracker_class_get_name (tracker_property_get_domain (property)));
+	}
+	field_name = tracker_property_get_name (property);
+
+	fts = tracker_property_get_fulltext_indexed (property);
+
+	/* read existing property values */
+	old_values = get_old_property_values (property, error);
+	if (error && *error) {
+		g_free (table_name);
+		return;
 	}
 
-	tracker_db_statement_bind_int (stmt, 0, resource_id);
+	string_to_gvalue (value, tracker_property_get_data_type (property), &gvalue);
 
-	switch (tracker_property_get_data_type (property)) {
-	case TRACKER_PROPERTY_TYPE_STRING:
-		tracker_db_statement_bind_text (stmt, 1, value);
-		break;
-	case TRACKER_PROPERTY_TYPE_INTEGER:
-		tracker_db_statement_bind_int (stmt, 1, atoi (value));
-		break;
-	case TRACKER_PROPERTY_TYPE_BOOLEAN:
-		tracker_db_statement_bind_int (stmt, 1, strcmp (value, "true") == 0);
-		break;
-	case TRACKER_PROPERTY_TYPE_DOUBLE:
-		tracker_db_statement_bind_double (stmt, 1, atof (value));
-		break;
-	case TRACKER_PROPERTY_TYPE_DATE:
-	case TRACKER_PROPERTY_TYPE_DATETIME:
-		tracker_db_statement_bind_int (stmt, 1, tracker_string_to_date (value));
-		break;
-	case TRACKER_PROPERTY_TYPE_RESOURCE:
-		object_id = ensure_resource_id (value);
-		tracker_db_statement_bind_int (stmt, 1, object_id);
-		break;
-	default:
-		g_assert_not_reached ();
+	if (!value_set_remove_value (old_values, &gvalue)) {
+		/* value not found */
+		g_value_unset (&gvalue);
+	} else {
+		cache_delete_value (table_name, field_name, &gvalue, multiple_values, fts);
 	}
 
-	tracker_db_statement_execute (stmt, NULL);
-	g_object_unref (stmt);
+	g_free (table_name);
 
 	/* also delete super property values */
 	super_properties = tracker_property_get_super_properties (property);
 	while (*super_properties) {
-		delete_metadata_decomposed (resource_id, *super_properties, value);
+		delete_metadata_decomposed (*super_properties, value, error);
 		super_properties++;
 	}
 }
 
+static void
+cache_delete_resource_type (TrackerClass *class)
+{
+	TrackerDBInterface *iface;
+	TrackerDBStatement *stmt;
+	TrackerDBResultSet *result_set;
+	TrackerProperty   **properties, **prop;
+	gboolean            found;
+	gint                i;
+
+	iface = tracker_db_manager_get_db_interface ();
+
+	found = FALSE;
+	for (i = 0; i < resource_buffer->types->len; i++) {
+		if (strcmp (g_ptr_array_index (resource_buffer->types, i), tracker_class_get_uri (class)) == 0) {
+			found = TRUE;
+		}
+	}
+
+	if (!found) {
+		/* type not found, nothing to do */
+		return;
+	}
+
+	/* retrieve all subclasses we need to remove from the subject
+	 * before we can remove the class specified as object of the statement */
+	stmt = tracker_db_interface_create_statement (iface,
+		"SELECT (SELECT Uri FROM \"rdfs:Resource\" WHERE ID = \"rdfs:Class_rdfs:subClassOf\".ID) "
+		"FROM \"rdfs:Resource_rdf:type\" INNER JOIN \"rdfs:Class_rdfs:subClassOf\" ON (\"rdf:type\" = \"rdfs:Class_rdfs:subClassOf\".ID) "
+		"WHERE \"rdfs:Resource_rdf:type\".ID = ? AND \"rdfs:subClassOf\" = (SELECT ID FROM \"rdfs:Resource\" WHERE Uri = ?)");
+	tracker_db_statement_bind_int (stmt, 0, resource_buffer->id);
+	tracker_db_statement_bind_text (stmt, 1, tracker_class_get_uri (class));
+	result_set = tracker_db_statement_execute (stmt, NULL);
+	g_object_unref (stmt);
+
+	if (result_set) {
+		do {
+			gchar *class_uri;
+
+			tracker_db_result_set_get (result_set, 0, &class_uri, -1);
+			cache_delete_resource_type (tracker_ontology_get_class_by_uri (class_uri));
+			g_free (class_uri);
+		} while (tracker_db_result_set_iter_next (result_set));
+
+		g_object_unref (result_set);
+	}
+
+	/* delete all property values */
+
+	properties = tracker_ontology_get_properties ();
+
+	for (prop = properties; *prop; prop++) {
+		gboolean            multiple_values, fts;
+		gchar              *table_name;
+		const gchar        *field_name;
+		GValueArray        *old_values;
+		gint                i;
+
+		if (tracker_property_get_domain (*prop) != class) {
+			continue;
+		}
+
+		multiple_values = tracker_property_get_multiple_values (*prop);
+		if (multiple_values) {
+			table_name = g_strdup_printf ("%s_%s",
+				tracker_class_get_name (class),
+				tracker_property_get_name (*prop));
+		} else {
+			table_name = g_strdup (tracker_class_get_name (class));
+		}
+		field_name = tracker_property_get_name (*prop);
+
+		fts = tracker_property_get_fulltext_indexed (*prop);
+
+		old_values = get_old_property_values (*prop, NULL);
+
+		for (i = old_values->n_values - 1; i >= 0 ; i--) {
+			GValue *old_gvalue;
+			GValue  gvalue = { 0 };
+
+			old_gvalue = g_value_array_get_nth (old_values, i);
+			g_value_init (&gvalue, G_VALUE_TYPE (old_gvalue));
+			g_value_copy (old_gvalue, &gvalue);
+
+			value_set_remove_value (old_values, &gvalue);
+			cache_delete_value (table_name, field_name, &gvalue, multiple_values, fts);
+		}
+
+		g_free (table_name);
+	}
+
+	cache_delete_row (class);
+
+	tracker_class_set_count (class, tracker_class_get_count (class) - 1);
+	if (delete_callback) {
+		delete_callback (resource_buffer->subject, RDF_PREFIX "type", tracker_class_get_uri (class), resource_buffer->types, delete_data);
+	}
+}
 
 void
 tracker_data_delete_statement (const gchar            *subject,
@@ -921,7 +1148,6 @@ tracker_data_delete_statement (const gchar            *subject,
 	TrackerClass       *class;
 	TrackerProperty    *field;
 	gint		    subject_id;
-	GPtrArray          *types;
 
 	g_return_if_fail (subject != NULL);
 	g_return_if_fail (predicate != NULL);
@@ -929,173 +1155,34 @@ tracker_data_delete_statement (const gchar            *subject,
 	g_return_if_fail (in_transaction);
 
 	subject_id = query_resource_id (subject);
-	
+
 	if (subject_id == 0) {
 		/* subject not in database */
 		return;
 	}
 
-	/* delete does not use update buffer, flush it */
-	tracker_data_update_buffer_flush ();
+	if (resource_buffer == NULL || strcmp (resource_buffer->subject, subject) != 0) {
+		/* switch subject */
+		resource_buffer = g_hash_table_lookup (update_buffer.resources, subject);
+	}
 
-	types = tracker_data_query_rdf_type (subject_id);
+	if (resource_buffer == NULL) {
+		/* subject not yet in cache, retrieve or create ID */
+		resource_buffer = g_slice_new0 (TrackerDataUpdateBufferResource);
+		resource_buffer->subject = g_strdup (subject);
+		resource_buffer->id = query_resource_id (resource_buffer->subject);
+		resource_buffer->fts_updated = FALSE;
+		resource_buffer->types = tracker_data_query_rdf_type (resource_buffer->id);
+		resource_buffer->predicates = g_hash_table_new_full (g_direct_hash, g_direct_equal, g_object_unref, (GDestroyNotify) g_value_array_free);
+		resource_buffer->tables = g_hash_table_new_full (g_str_hash, g_str_equal, g_free, (GDestroyNotify) cache_table_free);
 
-	/* TODO we need to retrieve all existing (FTS indexed) property values for
-	 * this resource to properly support incremental FTS updates
-	 * (like calling deleteTerms and then calling insertTerms)
-	 */
-	tracker_fts_update_init (subject_id);
+		g_hash_table_insert (update_buffer.resources, g_strdup (subject), resource_buffer);
+	}
 
 	if (object && g_strcmp0 (predicate, RDF_PREFIX "type") == 0) {
 		class = tracker_ontology_get_class_by_uri (object);
 		if (class != NULL) {
-			TrackerDBInterface *iface;
-			TrackerDBStatement *stmt;
-			TrackerDBResultSet *result_set;
-			TrackerProperty   **properties, **prop;
-			GString *projection = NULL;
-
-			iface = tracker_db_manager_get_db_interface ();
-
-			/* retrieve all subclasses we need to remove from the subject
-			 * before we can remove the class specified as object of the statement */
-			stmt = tracker_db_interface_create_statement (iface,
-				"SELECT (SELECT Uri FROM \"rdfs:Resource\" WHERE ID = \"rdfs:Class_rdfs:subClassOf\".ID) "
-				"FROM \"rdfs:Resource_rdf:type\" INNER JOIN \"rdfs:Class_rdfs:subClassOf\" ON (\"rdf:type\" = \"rdfs:Class_rdfs:subClassOf\".ID) "
-				"WHERE \"rdfs:Resource_rdf:type\".ID = ? AND \"rdfs:subClassOf\" = (SELECT ID FROM \"rdfs:Resource\" WHERE Uri = ?)");
-			tracker_db_statement_bind_int (stmt, 0, subject_id);
-			tracker_db_statement_bind_text (stmt, 1, object);
-			result_set = tracker_db_statement_execute (stmt, NULL);
-			g_object_unref (stmt);
-
-			if (result_set) {
-				do {
-					gchar *class_uri;
-
-					tracker_db_result_set_get (result_set, 0, &class_uri, -1);
-					tracker_data_delete_statement (subject, predicate, class_uri, NULL);
-					g_free (class_uri);
-				} while (tracker_db_result_set_iter_next (result_set));
-
-				g_object_unref (result_set);
-			}
-
-			properties = tracker_ontology_get_properties ();
-
-			for (prop = properties; *prop; prop++) {
-
-				if (tracker_property_get_domain (*prop) != class) {
-					continue;
-				}
-
-				if (!tracker_property_get_multiple_values (*prop)) {
-
-					if (tracker_property_get_fulltext_indexed (*prop)) {
-						if (!projection) {
-							projection = g_string_new ("");
-						} else {
-							g_string_append_c (projection, ',');
-						}
-
-						g_string_append_c (projection, '\'');
-						g_string_append (projection, tracker_property_get_name (*prop));
-						g_string_append (projection, "',");
-						g_string_append_c (projection, '"');
-						g_string_append (projection, tracker_property_get_name (*prop));
-						g_string_append_c (projection, '"');
-					}
-				} else {
-
-					if (tracker_property_get_fulltext_indexed (*prop)) {
-
-						/* Removing all fulltext properties from fts for 
-						 * none nrl:maxCardinality */
-
-						stmt = tracker_db_interface_create_statement (iface, "SELECT \"%s\" FROM \"%s_%s\" WHERE ID = ?",
-								tracker_property_get_name (*prop),
-								tracker_class_get_name (class),
-								tracker_property_get_name (*prop));
-						tracker_db_statement_bind_int (stmt, 0, subject_id);
-						result_set = tracker_db_statement_execute (stmt, NULL);
-						g_object_unref (stmt);
-
-						if (result_set) {
-							gchar *prop_object = NULL;
-
-							tracker_db_result_set_get (result_set, 0, &prop_object, -1);
-							g_object_unref (result_set);
-
-							if (prop_object) {
-								tracker_fts_update_text (subject_id, -1, prop_object);
-
-								g_free (prop_object);
-							}
-						}
-					}
-
-					/* multi-valued property, delete values from DB */
-					stmt = tracker_db_interface_create_statement (iface, "DELETE FROM \"%s_%s\" WHERE ID = ?",
-								tracker_class_get_name (class),
-								tracker_property_get_name (*prop));
-					tracker_db_statement_bind_int (stmt, 0, subject_id);
-					tracker_db_statement_execute (stmt, NULL);
-					g_object_unref (stmt);
-				}
-			}
-
-			/* delete single-valued properties for current class */
-
-			if (projection) {
-
-				/* Removing all fulltext properties from fts for
-				 * nrl:maxCardinality 1 */
-
-				stmt = tracker_db_interface_create_statement (iface, "SELECT %s FROM \"%s\" WHERE ID = ?",
-						projection->str,
-						tracker_class_get_name (class));
-				g_string_free (projection, TRUE);
-
-				tracker_db_statement_bind_int (stmt, 0, subject_id);
-				result_set = tracker_db_statement_execute (stmt, NULL);
-				g_object_unref (stmt);
-
-				if (result_set) {
-					gchar *prop_object = NULL;
-					gchar *prop_name = NULL;
-					guint columns, c;
-
-					do {
-						columns = tracker_db_result_set_get_n_columns (result_set);
-						for (c = 0; c < columns; c += 2) {
-							tracker_db_result_set_get (result_set, c, &prop_name, -1);
-							tracker_db_result_set_get (result_set, c + 1, &prop_object, -1);
-
-							if (prop_object && prop_name) {
-								tracker_fts_update_text (subject_id, -1, prop_object);
-							}
-
-							g_free (prop_object);
-							prop_object = NULL;
-							g_free (prop_name);
-							prop_name = NULL;
-
-						}
-					} while (tracker_db_result_set_iter_next (result_set));
-
-					g_object_unref (result_set);
-				}
-			}
-
-			stmt = tracker_db_interface_create_statement (iface, "DELETE FROM \"%s\" WHERE ID = ?",
-			                                              tracker_class_get_name (class));
-			tracker_db_statement_bind_int (stmt, 0, subject_id);
-			tracker_db_statement_execute (stmt, NULL);
-			g_object_unref (stmt);
-
-			/* delete rows from class tables */
-			delete_resource_type (subject_id, class);
-
-			tracker_class_set_count (class, tracker_class_get_count (class) - 1);
+			cache_delete_resource_type (class);
 		} else {
 			g_set_error (error, TRACKER_DATA_ERROR, TRACKER_DATA_ERROR_UNKNOWN_CLASS,
 				     "Class '%s' not found in the ontology", object);
@@ -1103,27 +1190,15 @@ tracker_data_delete_statement (const gchar            *subject,
 	} else {
 		field = tracker_ontology_get_property_by_uri (predicate);
 		if (field != NULL) {
-
-			delete_metadata_decomposed (subject_id, field, object);
-
-			if (object && tracker_property_get_fulltext_indexed (field)) {
-				tracker_fts_update_text (subject_id, -1, object);
-			}
+			delete_metadata_decomposed (field, object, error);
 		} else {
 			g_set_error (error, TRACKER_DATA_ERROR, TRACKER_DATA_ERROR_UNKNOWN_PROPERTY,
 				     "Property '%s' not found in the ontology", predicate);
 		}
-	}
 
-	update_buffer.fts_ever_updated = TRUE;
-
-	if (delete_callback) {
-		delete_callback (subject, predicate, object, types, delete_data);
-	}
-
-	if (types) {
-		g_ptr_array_foreach (types, (GFunc) g_free, NULL);
-		g_ptr_array_free (types, TRUE);
+		if (delete_callback) {
+			delete_callback (subject, predicate, object, resource_buffer->types, delete_data);
+		}
 	}
 }
 
@@ -1591,7 +1666,7 @@ tracker_data_commit_transaction (void)
 
 	in_transaction = FALSE;
 
-	tracker_data_update_buffer_flush ();
+	tracker_data_update_buffer_flush (NULL);
 
 	if (update_buffer.fts_ever_updated) {
 		tracker_fts_update_commit ();
@@ -1639,19 +1714,22 @@ format_sql_value_as_string (GString         *sql,
  * annotations (non-embedded/user metadata) stored about the resource.
  */
 void
-tracker_data_delete_resource_description (const gchar *uri)
+tracker_data_delete_resource_description (const gchar *uri, GError **error)
 {
 	TrackerDBInterface *iface;
 	TrackerDBStatement *stmt;
-	TrackerDBCursor    *cursor, *single_cursor, *multi_cursor;
+	TrackerDBResultSet *result_set, *single_result, *multi_result;
 	TrackerClass	   *class;
 	GString		   *sql;
 	TrackerProperty	  **properties, **property;
-	const gchar	   *class_uri, *value;
 	int		    i;
 	gboolean            first;
 	gint                resource_id;
 
+	/* We use result_sets instead of cursors here because it's possible
+	 * that otherwise the query of the outer cursor would be reused by the
+	 * cursors of the inner queries. */
+
 	resource_id = tracker_data_query_resource_id (uri);
 
 	iface = tracker_db_manager_get_db_interface ();
@@ -1660,18 +1738,23 @@ tracker_data_delete_resource_description (const gchar *uri)
 
 	stmt = tracker_db_interface_create_statement (iface, "SELECT (SELECT Uri FROM \"rdfs:Resource\" WHERE ID = \"rdf:type\") FROM \"rdfs:Resource_rdf:type\" WHERE ID = ?");
 	tracker_db_statement_bind_int (stmt, 0, resource_id);
-	cursor = tracker_db_statement_start_cursor (stmt, NULL);
+	result_set = tracker_db_statement_execute (stmt, NULL);
 	g_object_unref (stmt);
 
-	if (cursor) {
-		while (tracker_db_cursor_iter_next (cursor)) {
-			class_uri = tracker_db_cursor_get_string (cursor, 0);
+	if (result_set) {
+		do {
+			gchar *class_uri;
+
+			tracker_db_result_set_get (result_set, 0, &class_uri, -1);
 
 			class = tracker_ontology_get_class_by_uri (class_uri);
+
 			if (class == NULL) {
 				g_warning ("Class '%s' not found in the ontology", class_uri);
+				g_free (class_uri);
 				continue;
 			}
+			g_free (class_uri);
 
 			/* retrieve single value properties for current class */
 
@@ -1695,13 +1778,12 @@ tracker_data_delete_resource_description (const gchar *uri)
 				}
 			}
 
-			single_cursor = NULL;
+			single_result = NULL;
 			if (!first) {
 				g_string_append_printf (sql, " FROM \"%s\" WHERE ID = ?", tracker_class_get_name (class));
 				stmt = tracker_db_interface_create_statement (iface, "%s", sql->str);
 				tracker_db_statement_bind_int (stmt, 0, resource_id);
-				single_cursor = tracker_db_statement_start_cursor (stmt, NULL);
-				tracker_db_cursor_iter_next (single_cursor);
+				single_result = tracker_db_statement_execute (stmt, NULL);
 				g_object_unref (stmt);
 			}
 
@@ -1723,15 +1805,18 @@ tracker_data_delete_resource_description (const gchar *uri)
 				}
 
 				if (!tracker_property_get_multiple_values (*property)) {
+					gchar *value;
+
 					/* single value property, value in single_result_set */
 
-					value = tracker_db_cursor_get_string (single_cursor, i++);
+					tracker_db_result_set_get (single_result, i++, &value, -1);
 
 					if (value) {
 						tracker_data_delete_statement (uri, 
 						                               tracker_property_get_uri (*property), 
 						                               value, 
-						                               NULL);
+						                               error);
+						g_free (value);
 					}
 
 				} else {
@@ -1748,22 +1833,25 @@ tracker_data_delete_resource_description (const gchar *uri)
 
 					stmt = tracker_db_interface_create_statement (iface, "%s", sql->str);
 					tracker_db_statement_bind_int (stmt, 0, resource_id);
-					multi_cursor = tracker_db_statement_start_cursor (stmt, NULL);
+					multi_result = tracker_db_statement_execute (stmt, NULL);
 					g_object_unref (stmt);
 
-					if (multi_cursor) {
-						while (tracker_db_cursor_iter_next (multi_cursor)) {
+					if (multi_result) {
+						do {
+							gchar *value;
 
-							value = tracker_db_cursor_get_string (multi_cursor, 0);
+							tracker_db_result_set_get (multi_result, 0, &value, -1);
 
 							tracker_data_delete_statement (uri, 
 							                               tracker_property_get_uri (*property), 
 							                               value,
 							                               NULL);
 
-						}
+							g_free (value);
 
-						g_object_unref (multi_cursor);
+						} while (tracker_db_result_set_iter_next (multi_result));
+
+						g_object_unref (multi_result);
 					}
 
 					g_string_free (sql, TRUE);
@@ -1771,12 +1859,12 @@ tracker_data_delete_resource_description (const gchar *uri)
 			}
 
 			if (!first) {
-				g_object_unref (single_cursor);
+				g_object_unref (single_result);
 			}
 
-		}
+		} while (tracker_db_result_set_iter_next (result_set));
 
-		g_object_unref (cursor);
+		g_object_unref (result_set);
 	}
 }
 
@@ -1801,6 +1889,10 @@ tracker_data_update_sparql (const gchar  *update,
 	if (*error) {
 		tracker_data_update_buffer_clear ();
 		tracker_db_interface_execute_query (iface, NULL, "ROLLBACK TO sparql");
+
+		if (rollback_callback) {
+			rollback_callback (commit_data);
+		}
 	}
 
 	tracker_db_interface_execute_query (iface, NULL, "RELEASE sparql");
